;;;; -*- Mode: Lisp -*-
;;;; Extracted from Fare's personal utilities

(defpackage #:fare-matcher-utils
  (:use #:common-lisp)
  (:export #:propmacros
           #:make-single-arg-form
           #:define-abbrevs
           #:length=-p
           #:length=n-p
           #:length<-p
           #:length<n-p
           #:length<=-p
           #:length<=n-p
           #:length>n-p
           #:length>=n-p
           #:conc-gensym
           #:aif
           #:combinationp
           #:literalp
           #:mapcar2
           #:mvbind
           #:plist->alist
           #:single-arg
           #:it))

(in-package #:fare-matcher-utils)

(eval-when (:compile-toplevel :load-toplevel :execute)
  (defun error-behaviour (e &rest r)
    "generic way to specify behaviour in exceptional situations"
    (cond
      ((functionp e) (apply e r))
      ((stringp e) (apply #'error e r))
      (t e)))

  (defun stuff->string (stuff)
    "transform some stuff into a string"
    (typecase stuff
      (character (make-string 1 :initial-element stuff))
      (string stuff)
      (symbol (symbol-name stuff))
      (t (format nil "~A" stuff))))

  (defun plist->alist (pairs)
    "transform a list of consecutive pairs into an alist"
    (assert (and (listp pairs) (evenp (list-length pairs))))
    (if (null pairs) nil
        (acons (car pairs) (cadr pairs) (plist->alist (cddr pairs)))))

  (defun conc-string (&rest rest)
    "make a string by concatenating stuff"
    (apply #'concatenate 'string (mapcar #'stuff->string rest)))
  
  (defun conc-gensym (&rest rest)
    "make a symbol by concatenating stuff"
    (make-symbol (apply #'conc-string rest)))

  (defun conc-symbol (&rest rest)
    "make a symbol by concatenating stuff"
    (apply #'conc-symbol-in nil rest))

  (defun conc-symbol-in (package &rest rest)
    "make a symbol by concatenating stuff"
    (intern (apply #'conc-string rest) (or package *package*)))
  
  (defmacro let1 (var val &body body) `(let ((,var ,val)) ,@body))
  
  (defun make-predicate-symbol (s)
    (let1 n (conc-string s)
          (conc-symbol n (if (every #'alphanumericp n) :p :-p)))))

(defmacro define-abbrevs (&rest abbrevs)
  "Declare a series of abbreviations."
  `(progn ,@(mapcar #'(lambda (x)
			`(defmacro ,(car x) (&rest rest) `(,',(cdr x) ,@rest)))
		    (plist->alist abbrevs))))

(define-abbrevs
  mvbind multiple-value-bind
  mvcall multiple-value-call
  mvlist multiple-value-list
  mvprog1 multiple-value-prog1
  mvsetq multiple-value-setq)

(defun mapcar2 (fun list)
  "simple mapcar for functions returning 2 values,
returning the two lists of the values returned by the function."
  (if (null list) (values nil nil)
    (mvbind (x y) (funcall fun (car list))
       (mvbind (xl yl) (mapcar2 fun (cdr list))
	  (values (cons x xl) (cons y yl))))))

(defun length=-p (x y)
  ;(= (length x) (length y))
  (loop
    (let ((nx (null x)) (ny (null y)))
      (if nx (return ny)
        (if ny (return nil)
          (setf x (cdr x) y (cdr y)))))))
(defun length=n-p (x n)
  ;(= (length x) n)
  (loop
    (let ((nx (null x)) (zn (zerop n)))
      (if nx (return zn)
        (if zn (return nil)
          (setf x (cdr x) n (1- n)))))))
(defun length<-p (x y)
  ;(= (length x) (length y))
  (loop
    (if (null y) (return nil)
      (if (null x) (return t)
        (setf x (cdr x) y (cdr y))))))
(defun length<n-p (x n)
  ;(< (length x) n)
  (loop
    (if (zerop n) (return nil)
      (if (null x) (return t)
        (setf x (cdr x) n (1- n))))))
(defun length<=-p (x y)
  (not (length<-p y x)))
(defun length<=n-p (x n)
  (length<n-p x (1+ n)))
(defun length>n-p (x n)
  (not (length<=n-p x n)))
(defun length>=n-p (x n)
  (not (length<n-p x n)))

(defmacro aif (test then &optional else)
  `(let ((it ,test)) (if it ,then ,else)))

(defmacro mapmacro (fun forms &optional (head '(progn)) (tail '()))
  `(,@head ,@(mapcar #'(lambda (x) `(,fun ,x)) forms) ,@tail))
(defmacro propmacro (name)
  `(defmacro ,name (obj) `(get ,obj ',',name)))
(defmacro propmacros (&rest names)
  `(mapmacro propmacro ,names))

(defun single-arg-form-p (tag x &optional on-error)
  "checks whether X is an instance of a something made with a maker
defined by (MAKE-SINGLE-ARG-FORM TAG). If ON-ERROR is defined, and
X looks like it is such an instance but is malformed, then ON-ERROR
is invoked as an ERROR-BEHAVIOUR."
  (and (consp x) (eq tag (car x))
       (or (and (consp (cdr x)) (null (cddr x)))
	   (error-behaviour on-error tag x))))

(defun single-arg (x) (cadr x))
(defmacro make-single-arg-form (name &optional
				     (maker (conc-symbol :make- name))
				     (recognizer (make-predicate-symbol name))
				     (tag name))
  `(progn
     (defun ,maker (x) (list ',tag x))
     (defun ,recognizer (x &optional on-error)
       (single-arg-form-p ',tag x on-error))))
(defun proper-list-p (list)
  (values (ignore-errors (list-length list))))
(defun combinationp (x &optional on-error)
  "Is the form X a combination of a head and arguments,
to be evaluated as function call, macro call, or special form?"
  (and (consp x)
       (if on-error
	   (or (proper-list-p x) (error-behaviour on-error x))
	 t)))
(defun literalp (x)
  "predicate that tells whether X the source form for a literal expression."
  (typep x '(or boolean number character string keyword)))
